/**********************************************************************************/
/***	projekt:	Generator stromov, bakalarska praca							***/
/***	autor:		Tomas Pafco, xpafco00@stud.fit.vutbr.cz						***/
/***	rok:		2006/2007													***/
/***																			***/
/***	subor:		TreeTest.cpp												***/
/***	popis:		funkcia main(), funkcie generujuce zakladnu strukturu 		***/
/***				stromu														***/
/***																			***/
/***	poznamky:	vytvorene vo Visual Studiu 2003 pod WinXP					***/
/***																			***/
/**********************************************************************************/


#include <osg/Group>
#include <osg/StateSet>
#include <osg/ShadeModel>
#include <osg/Texture2D>
#include <osg/Projection>

#include <osgDB/FileUtils>
#include <osgProducer/Viewer>
#include <osgUtil/Optimizer>
#include <math.h>

#include "parser.h"
#include "randomc.h"
#include "geometria.h"
#include "povrch.h"
#include "listy.h"
#include "tex.h"
#include "hud.h"
#include "ovladanie.h"

using namespace std;


// pocet a konf. subory stromov ktore sa budu generovat
#define SPOCET 7					



// ................................................................................
// ........................... implicitne nastavenie parametrov ...................
// ................................................................................
//
tparam Parametre;

void ParamInit()
{
	Parametre.vyska_stromu =	3;			// vyska stromu
	Parametre.max_rozptyl_kmen = 0.1;		// maximum rozptylu kontrolnych bodov kmena
	Parametre.sirka_vyska =		0.9;		// pomer: sirka stromu / vyska stromu
	Parametre.koruna_kmen =		0.75;		// pomer: vyska koruny / vyska kmena
	Parametre.seg_vet =			4;			// segmentacia vetiev (realne cislo)
	Parametre.seg_kmen =		1;			// segmentacia hmena (integer!)
	Parametre.alfa_kor =		0.2;		// korekcia rozsahov nahodnych hodnot smeroveho uhla vetiev
	Parametre.v_rozptyl =		0.3; 		// miera rozptylu kontrolnych bodov vetvy (float)
	Parametre.krv_pomer =		0.02;		// pomer: spodny polomer kmena / vyska stromu
	Parametre.vrv_pomer =		0.012;		// pomer: spodny polomer vetvy / dlzka vetvy
	Parametre.lsph_polomer =	0.25;		// vzdialenost listov od stredu (sfericka metoda)
	Parametre.lsph_pocet =		0;			// pocet listov v zhluku (sfericka metoda)
	Parametre.lbil_dlzka =		0.4;		// dlzka listoveho bilbordu (bilbordova metoda)
	Parametre.lbil_sirka =		0.4;		// sirka listoveho bilbordu (bilbordova metoda)
	Parametre.lbil_uhol =		0.7;		// vzajomny uhol listoveho bilbordu a vetvy
	Parametre.l_velkost =		0.3;		// velkost listu (sfericka metoda)
	Parametre.ltex_pocet =		0;			// pocet roznych textur listu
	Parametre.sph_listy =		false;		// 1 - zobrazit listy - sfef. metoda, 0 - listy nezobrazovat
	Parametre.bil_listy =		false;		// 1 - zobrazit listy - bil. metoda, 0 - listy nezobrazovat
	Parametre.okr_koruna =		true;		// 1 - okruhla koruna, 0 - kuzelova
	Parametre.v1G_pocet =		1;			// pocet vetvi 1G vyrastajucich z jedneho k_bodu
	Parametre.bil_transluc =	0.0;		// priesvitnost bilbordu (0 - 1)

	Parametre.k_subor = "kora.png";			// subor s texturou kory
	Parametre.hTexMul =		1;				// horizontalny nasobitel textury kory (integer)
	Parametre.vTexMul =		1;				// vertikalny nasobitel textury kory (integer)

	// vetvy 1. generacie
	Parametre.v1G_uhol =	0.6;			// vzajomy uhol kmena a smer. vektora vetvy (rad)
	Parametre.uhp =			0.0;			// prirastok uhla kmena a vetvy
	Parametre.nextG_pomer =	1.0;			// pomer: usek vetvy z kt. vyrastaju vetvy nasledujucej generacie / cela vetva
	Parametre.v1G_grav =	0.12;			// vplyv gravitacie
	Parametre.v1G_sun =		0.07;			// miera zakrivenia smerom hore

	// vetvy 2. generacie
	Parametre.v2gener =		false;			// 1 - zobrazit vetvy 2. genracie, 0 - nezobrazit ...
	Parametre.v2G_uhol =	0.7;			// vzajomy uhol vetvy a rodicovskej vetvy (rad)
	Parametre.v2G_grav =	0.0;			// vplyv gravitacie
	Parametre.v2G_sun =		0.3;			// miera zakrivenia smerom hore
	Parametre.v2G_max =		2.0;			// maximalna dlzka vetvy

	// vetvy 3. generacie
	Parametre.v3gener =		false;			// 1 - zobrazit vetvy 2. genracie, 0 - nezobrazit ...
	Parametre.v3G_uhol =	0.9;			// vzajomy uhol vetvy a rodicovskej vetvy (rad)
	Parametre.v3G_grav =	0.0;			// vplyv gravitacie
	Parametre.v3G_sun =		0.0;			// miera zakrivenia smerom hore
	Parametre.v3G_max =		1.0;			// maximalna dlzka vetvy

	Parametre.detail =		1;				// glob. detaily: 3-najvyssie, 2-stredne, 1-nizke
	
}


const float pi = 3.14159265;

// ::::::::::::::::::::::::::: funkcia Kmen ::::::::::::::::::::::::::::::::::::::
// funkcia vytvori kmen stromu
//
void Kmen(osg::Group & root, TRandomMersenne & rg, osg::Vec4Array & kmen_body, 
		   float vyska)
{
	int i, p;
	int v_int;
	float rKor, polomer;		
	float ciast_rozptyl;
	float x, y, r, dx, dy;
	
	ciast_rozptyl = Parametre.max_rozptyl_kmen / vyska;
	v_int = (int)vyska * Parametre.seg_kmen;

	dx = 0;		// nahodny prirastok v smere x
	dy = 0;		// nahodny prirastok v smere y
	r = 0;		// limit nahodneho prirastku 
	x = 0;		// poloha bodu na x
	y = 0;		// poloha bodu na y
	rKor = 1;	// uprava polomeru kmena
	for (i = 0; i < v_int; i++)
	{
		polomer = vyska * Parametre.krv_pomer * rKor;
		if (polomer < 0.001)
			polomer = 0.0;					// polomer "spicky" je 0
		rKor = rKor - (1.0/(v_int - 1));
		
		kmen_body.push_back(osg::Vec4f(x, y, ((float)i) / Parametre.seg_kmen, polomer));
		dx = XRandom(-r, r, rg);
		dy = XRandom(-r, r, rg);
		x = x + dx;
		y = y + dy;
		r = r + ciast_rozptyl;
	}

	// povrch
	KTelo(root, kmen_body, Parametre.detail, Parametre.hTexMul, Parametre.vTexMul, 0);
	
	// listy
	// sefericka metoda
	if (Parametre.sph_listy)
		LSphSkupina(root, kmen_body.back().x(), kmen_body.back().y(), kmen_body.back().z(),
					Parametre.lsph_polomer, Parametre.lsph_pocet, Parametre.l_velkost, rg);

	// bilbordova metoda
	if (Parametre.bil_listy)
	{
		p = kmen_body.getNumElements();
		LBilbord(root, rg, osg::Vec4(kmen_body.back().x(), kmen_body.back().y(), kmen_body.back().z(), 0), 
			osg::Vec3(	kmen_body[p-2].x() - kmen_body[p-1].x(),
						kmen_body[p-2].y() - kmen_body[p-1].y(),
						kmen_body[p-2].z() - kmen_body[p-1].z()	),
			0, Parametre.lbil_dlzka, Parametre.lbil_sirka, 0, Parametre.bil_transluc);
	}
	//*
}

// ::::::::::::::::::::::::::: funkcia Vetva ::::::::::::::::::::::::::::::::::::::
// funkcia vykresli vetvu podla zadanych parametrov
//	g_vplyv simuluje vplyv gravitacie
//  s_vplyv urcuje zahrivenie konca vetvy smerom hore
//
void Vetva (osg::Group & root, 
			TRandomMersenne & rg, 
			osg::Vec4 start_bod, 
			osg::Vec4Array & 
			K_body_koruny,
			osg::Vec3 k_vektor, 
			float uhol, 
			float dlzka, 
			float g_vplyv, 
			float s_vplyv, 
			int kvadrant, 
			float ng_pomer, 
			int gen)
{
	float x, dx, alfa;
	float y, dy, beta;
	float z, dz;
	int i, p;
	
	float polomer = 0;												// polomer vetvy v urcitom bode
	float rKor = 1;													// korekcia polomeru
	float vrv_pom;													// pomer polomer vetvy/dlzka
	float ciast_dlzka = dlzka / (dlzka * Parametre.seg_vet);					// dlzka segmentu vetvy
	int seg_pocet = (int)(dlzka / ciast_dlzka + 0.5);				// pocet segmentov vetvy
	
	osg::Vec4Array * K_body_vetvy = new osg::Vec4Array;		// k-body ktore sa budu zobrazovat
	osg::Vec3 pomocny_bod(0, 0, 0);							// pomocny pri vypocte rozptylu

	
	beta = uhol;			// pociatocny uhol odklonenia od rodicovskej vetvy, resp. kmena
	
	if (kvadrant == 1)
	{
		// vetva v 1. kvadrante
		alfa = XRandom(0 + Parametre.alfa_kor, pi/2 - Parametre.alfa_kor, rg);
	}
	else if (kvadrant == 2)
	{
		// vetva v 2. kvadrante
		alfa = XRandom(pi/2 + Parametre.alfa_kor, pi - Parametre.alfa_kor, rg);
	}
	else if (kvadrant == 3)
	{
		// vetva v 3. kvadrante
		alfa = XRandom(pi + Parametre.alfa_kor, (3*pi)/2 - Parametre.alfa_kor, rg);
	}
	else if (kvadrant == 4)
	{
		// vetva v 4. kvadrante
		alfa = XRandom((3*pi)/2 + Parametre.alfa_kor, 2*pi - Parametre.alfa_kor, rg);
	}
	else
		// na kvadrante nezalezi
		alfa = XRandom(0, 2*pi, rg);

	
	
	x = start_bod.x();					//
	y = start_bod.y();					//	suradnice pociatocneho bodu vetvy
	z = start_bod.z();					//
	
	dx = ciast_dlzka * sin(beta) * cos(alfa);			//
	dy = ciast_dlzka * sin(beta) * sin(alfa);			// smerovy vektor vetvy (vzhladom na os z)
	dz = ciast_dlzka * cos(beta);						//

	// transformacia smeroveho vektoru
	SMTrans(x, y, z, dx, dy, dz, k_vektor);
	
	// VYGENEROVANIE VETVY
	vrv_pom = Parametre.vrv_pomer;
	for (i = 0; i <= seg_pocet; i++)
	{
		polomer = dlzka * vrv_pom * rKor;
		
		// kontrola ci polomer vetvy nieje vacsi ako polomer rodicovskej
		if (polomer >= start_bod.w())
		{
			// vypocita sa novy pomer
			vrv_pom = start_bod.w()/(dlzka*rKor);
			// novy polomer
			polomer = dlzka * vrv_pom * rKor;
		}

		if (polomer < 0.0001)
			polomer = 0.0;					// polomer "spicky" je 0
		rKor = rKor - (1.0/seg_pocet);
		

		
		if ((i > 0) && (i > ((float)seg_pocet - ((float)seg_pocet * ng_pomer))))
		// bod je platny ako pociatocny bod pre vyrastanie vetvy dalsej generacie
		{
			// ulozenie bodu do pola bodov koruny, 
			// tieto body sluzia ako pociatocne body pre vyrastanie dalsej generacie vetiev
			K_body_koruny.push_back(osg::Vec4(x, y, z, polomer));
		}
		
		// ulozenie bodu do pola bodov vetvy, pole sluzi na zobrazenie vetvy
		K_body_vetvy->push_back(osg::Vec4(x, y, z, polomer));
		
		// transformacia smeroveho vektoru, zohladni sa vplyv "gravitacie" a zakrivenie smerom hore
		GTrans(dx, dy, dz, g_vplyv);
		
		pomocny_bod.set(x + dx, y + dy, z + dz);		// novy bod (aktualne suradnice + smer.vektor (prirastok))
		RozptylV(pomocny_bod, x, y, z, Parametre.v_rozptyl, rg);	// transformacia bodu - nahodny "rozptyl"
		x = pomocny_bod.x();							// -.
		y = pomocny_bod.y();							// -|- suradnice transformovaneho bodu
		z = pomocny_bod.z();							// -'

		// od polovice vetvy sa uplatni zakrivenie smerom hore
		if (i > seg_pocet/2)
				g_vplyv = g_vplyv - s_vplyv;
					
	}
	
	// vytvorenie povrchu
	KTelo(root, *K_body_vetvy, Parametre.detail, Parametre.hTexMul, Parametre.vTexMul, gen);
	
	// listy
	// sfericka metoda
	if (Parametre.sph_listy)
		LSphSkupina(root, K_body_vetvy->back().x(), K_body_vetvy->back().y(), K_body_vetvy->back().z(),
					Parametre.lsph_polomer, Parametre.lsph_pocet, Parametre.l_velkost, rg);
	
	// bilbordova metoda
	if (Parametre.bil_listy)
	{
		p = K_body_vetvy->getNumElements();
		LBilbord(root, rg, osg::Vec4(K_body_vetvy->back().x(), K_body_vetvy->back().y(), K_body_vetvy->back().z(), 0), 
			osg::Vec3(	(*K_body_vetvy)[p-2].x() - (*K_body_vetvy)[p-1].x(),
						(*K_body_vetvy)[p-2].y() - (*K_body_vetvy)[p-1].y(),
						(*K_body_vetvy)[p-2].z() - (*K_body_vetvy)[p-1].z()	),
			pi, Parametre.lbil_dlzka, Parametre.lbil_sirka, 0, Parametre.bil_transluc);
	}
	
	// uvolnenie pamati
	K_body_vetvy->clear();
	//*
}


// ::::::::::::::::::::::::::: funkcia Koruna :::::::::::::::::::::::::::::::::::::
// funckia vytvori korunu stromu
//
void Koruna (osg::Group & root, TRandomMersenne & rg, osg::Vec4Array & A_body, 
			 float vyska)
{
	float rKoruny;			// polomer koruny
	int A_pocet;			// pocet vychodiskovych bodov pre vetvy
	int K_pocet;			// pocet kontrolnych bodov kmena
	int V_pocet;			// pocet pociatocnych bodov pre vetvy 2. gneneracie
	int i, j, p;			// indexy
	int kvadrant = 0;		// kvadrant v ktorom bude vyrastat nasledujuca vetva (0 -> lubovolny)
	float x;				// pomocna pri vypocte tvaru koruny
	float a, b, c, r;		// pomocne, vypocet k_vektoru
	
	float dlzka;			// dlzka konkretnej vetvy
	float max_dlzka;		// maximalna dlzka vetvy vyssej generacie
	float vzdialenost;		// vzdialenost dvoch bodov
	float uh_prir = 0.0;	// prirastok uhla kmena a vetvy 1. generacie
	osg::Vec4 start_bod;	// pociatocny bod vetvy
	osg::Vec3 termin;		// koncovy bod vetvy
	osg::Vec3 k_vektor;		// vektor, vzhladom ku ktoremu sa bude pocitat pociat. smernicovy vektor vetvy					
	osg::Vec4Array* K_body_koruny = new osg::Vec4Array;		// pole pociatocnych bodov vetiev

	rKoruny = ((vyska * Parametre.sirka_vyska) / 2);
	

	// .... faza 1 - hlavne vetvy .................................................
	A_pocet = (int)(vyska * Parametre.koruna_kmen * Parametre.seg_kmen);
	K_pocet = A_body.getNumElements();
	float *f = new float[A_pocet+1];
	
	// pole definujuce tvar koruny
	x = 0.1;
	for (i = 0; i <= A_pocet; i++)
	{
		if (Parametre.okr_koruna)
		// okruhla koruna
		{
			f[i] = sin(x * pi);
		
			if (f[i] > 0.9)
				f[i] = 1;

			if (f[i] < 0.33)
				f[i] = 0.33;
		
			x = x + (1 / (float)A_pocet)*0.6;
		}
		else 
		// kuzelova koruna
		{
			f[i] = x;
		
			if (f[i] > 0.9)
				f[i] = 0.8;

			if (f[i] < 0.15)
				f[i] = 0.10;
		
			x = x + (1 / (float)A_pocet);
		}
	}
	
	kvadrant = 0;
	k_vektor.set(0, 0, 1);			// kontrolny vektor rovnobezny s osou z
	j = 0;
	for (i = (K_pocet - 2); i >= (K_pocet - A_pocet - 1); i--)
	{
		dlzka = rKoruny * f[j];
		uh_prir = uh_prir + Parametre.uhp;
		j++;
		
		start_bod.set(A_body[i].x(), A_body[i].y(), A_body[i].z(), A_body[i].w());
		
		
		for (p = 0; p < Parametre.v1G_pocet; p++)
		{
			kvadrant++;
			if (kvadrant > 4)
				kvadrant = 1;

			Vetva(root, rg, start_bod, *K_body_koruny, k_vektor, Parametre.v1G_uhol + uh_prir, dlzka, 
				Parametre.v1G_grav, Parametre.v1G_sun, kvadrant, Parametre.nextG_pomer, 1);
		}
		
	}
	// .... (1) ...................................................................

	V_pocet = K_body_koruny->getNumElements();
	max_dlzka = Parametre.v2G_max;
	
	// .... faza 2 ................................................................
	//
	if (Parametre.v2gener)
	{
		for (i = 0; i < V_pocet; i++)
		{
			start_bod.set((*K_body_koruny)[i].x(), (*K_body_koruny)[i].y(), (*K_body_koruny)[i].z(), (*K_body_koruny)[i].w());
		
			if (start_bod.w() != 0)
			{
				a = (*K_body_koruny)[i + 1].x() - (*K_body_koruny)[i].x();
				b = (*K_body_koruny)[i + 1].y() - (*K_body_koruny)[i].y();
				c = (*K_body_koruny)[i + 1].z() - (*K_body_koruny)[i].z();
				k_vektor.set(a, b, c);
			
				kvadrant++;
				if (kvadrant > 4)
					kvadrant = 1;
			
				// najdenie koncoveho bodu
				j = i;
				r = (*K_body_koruny)[j].w();
				while (r != 0.0)
				{
					j++;
					r = (*K_body_koruny)[j].w();
				}
				termin.set((*K_body_koruny)[j].x(), (*K_body_koruny)[j].y(), (*K_body_koruny)[j].z());
			
				// urcenie vzdialenosti start_bodu od koncoveho
				a = termin.x() - start_bod.x();
				b = termin.y() - start_bod.y();
				c = termin.z() - start_bod.z();
				vzdialenost = sqrt(a*a + b*b + c*c);
			
				// uprava dlzky
				dlzka = max_dlzka;
				if (vzdialenost < max_dlzka)
					dlzka = vzdialenost;

				Vetva(root, rg, start_bod, *K_body_koruny, k_vektor, Parametre.v2G_uhol, dlzka, 
					Parametre.v2G_grav, Parametre.v2G_sun, kvadrant, 1.0, 2);
			
				if (!Parametre.bil_listy)
				// odstranenie bodu z ktoreho vyrasta vetva
				// v pripade ze je pouzita bilbordova metoda, bod sa neodstrani,
				// aby mohol byt pociatocnym bodom pre bilbord
				//
				{
					K_body_koruny->erase(K_body_koruny->begin() + i);
				}
			
			}
		}
	}			// faza 2

	V_pocet = K_body_koruny->getNumElements();
	max_dlzka = Parametre.v3G_max;
	
	// .... faza 3 ................................................................
	//
	if (Parametre.v3gener)
	{
		kvadrant = 0;
		for (i = 0; i < V_pocet; i++)
		{
			start_bod.set((*K_body_koruny)[i].x(), (*K_body_koruny)[i].y(), (*K_body_koruny)[i].z(), (*K_body_koruny)[i].w());
		
			if (start_bod.w() != 0)
			{
				a = (*K_body_koruny)[i + 1].x() - (*K_body_koruny)[i].x();
				b = (*K_body_koruny)[i + 1].y() - (*K_body_koruny)[i].y();
				c = (*K_body_koruny)[i + 1].z() - (*K_body_koruny)[i].z();
				k_vektor.set(a, b, c);
			
				kvadrant++;
				if (kvadrant > 4)
					kvadrant = 1;
			
				// najdenie koncoveho bodu
				j = i;
				r = (*K_body_koruny)[j].w();
				while (r != 0.0)
				{
					j++;
					r = (*K_body_koruny)[j].w();
				}
				termin.set((*K_body_koruny)[j].x(), (*K_body_koruny)[j].y(), (*K_body_koruny)[j].z());
			
				// urcenie vzdialenosti start_bodu od koncoveho
				a = termin.x() - start_bod.x();
				b = termin.y() - start_bod.y();
				c = termin.z() - start_bod.z();
				vzdialenost = sqrt(a*a + b*b + c*c);
			
				// uprava dlzky
				dlzka = max_dlzka;
				if (vzdialenost < max_dlzka)
					dlzka = vzdialenost;

				Vetva(root, rg, start_bod, *K_body_koruny, k_vektor, Parametre.v3G_uhol, dlzka, 
					Parametre.v3G_grav, Parametre.v3G_sun, kvadrant, 1.0, 3);
			
				
				// odstranenie bodu z ktoreho vyrasta vetva
				K_body_koruny->erase(K_body_koruny->begin() + i);
				
			
			}
		}
	}			// faza 3

	// ............................................................................
	// listove bilbordy
	//
	if (Parametre.bil_listy)
	{
		V_pocet = K_body_koruny->getNumElements();
		kvadrant = 0;
		for (i = 0; i < V_pocet; i++)
		{
			start_bod.set((*K_body_koruny)[i].x(), (*K_body_koruny)[i].y(), (*K_body_koruny)[i].z(), (*K_body_koruny)[i].w());
		
			if (start_bod.w() != 0)
			{
				a = (*K_body_koruny)[i + 1].x() - (*K_body_koruny)[i].x();
				b = (*K_body_koruny)[i + 1].y() - (*K_body_koruny)[i].y();
				c = (*K_body_koruny)[i + 1].z() - (*K_body_koruny)[i].z();
				k_vektor.set(a, b, c);
			
				kvadrant++;
				if (kvadrant > 4)
					kvadrant = 1;
			
				
				LBilbord(root, rg, start_bod, k_vektor, 
					Parametre.lbil_uhol, Parametre.lbil_dlzka, Parametre.lbil_sirka, kvadrant, Parametre.bil_transluc);
				
			}
		}
	}

	// uvolnenie pamati
	K_body_koruny->clear();
	K_body_koruny->unref();
	delete[] f;

}

// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ::::::::::::::::::::::::::: HLAVNY PROGRAM :::::::::::::::::::::::::::::::::::::
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main(int argc, char **argv)
{
	myKeyboardEventHandler* KeyOvladanie = new myKeyboardEventHandler();
	osg::Projection* osd = new osg::Projection;
	osg::Group* superbase = new osg::Group;
	osgProducer::Viewer viewer;
	char* subory[] = {	"_zima",
						"ceresna_jar",
						"javor",
						"smrek",
						"borovica",
						"vrba",
						"_jesen"};
	char* stromy[] = {"zimny strom","jarna ceresna","javor","smrek","borovica","vrba","jesenny strom"};
	osgUtil::Optimizer optimizer;
	osg::Vec4Array* A_body = new osg::Vec4Array;
	int i, e;
	
	// cesty k suborom
	osgDB::FilePathList pathList = osgDB::getDataFilePathList();
	pathList.push_back("..\\Textury\\");
	pathList.push_back("..\\Fonty\\");
	pathList.push_back("..\\Konf_subory\\");
	osgDB::setDataFilePathList(pathList);
	
	
	// inicializacia paramterov (default)
	ParamInit();

	// vytvorenie a inicializacia generatora nahodnych cisel
	int32 seed = time(0);
	TRandomMersenne* rg = new TRandomMersenne(seed);

	// vytvorenie OSD
	osd = OnSD();
	
	cout << "genrovanie stromov, prosim cakajte ..." << endl;

	// GENEROVANIE STROMOV
	//
	for (i = 0; i < SPOCET; i++)
	{
		osg::ShadeModel* shModel = new osg::ShadeModel;
		osg::StateSet* stSet = new osg::StateSet;
		osg::Texture2D* kora;
		osg::Group* root = new osg::Group();
		
		// vyprazdnenie pola textur a pola zakladnych bodov
		textury.clear();	
		A_body->clear();
		
		cout << "generuje sa " << stromy[i] << " ... ";

		// nacitanie parametrov
		e = GetParam(Parametre, subory[i]);
		if (e != 0)
			// nepodarilo sa otvorit konfiguracny subor
			continue;

		// nacitanie suborov s texturami
		kora = KoraL(Parametre.k_subor);
		e = TexLoud(Parametre.ltex_pocet, Parametre.tex_subory);
		if ((kora == NULL) || (e < 0))
		{
			// pokracovanie dalsim stromom, ak sa nepodarilo nacitat potrebne textury
			continue;
		}

		// vytvorenie stromu
		Kmen(*root, *rg, *A_body, Parametre.vyska_stromu);
		Koruna(*root, *rg, *A_body, Parametre.vyska_stromu);

		// nastavenie tienovania a textury
		shModel->setMode(osg::ShadeModel::Mode::SMOOTH);
		stSet->setAttribute(shModel);
		kora->setWrap(osg::Texture::WrapParameter::WRAP_R, osg::Texture::WrapMode::REPEAT);
		kora->setWrap(osg::Texture::WrapParameter::WRAP_S, osg::Texture::WrapMode::REPEAT);
		kora->setWrap(osg::Texture::WrapParameter::WRAP_T, osg::Texture::WrapMode::REPEAT);
		stSet->setTextureAttributeAndModes(0,kora,osg::StateAttribute::ON);
		root->setStateSet(stSet);

		// pridanie stromu do prepinaca
		base->addChild(root);

		cout << " OK" << endl;
	}
	
	base->setSingleChildOn(st_index);
	osd_switch->addChild(osd);
	superbase->addChild(base);
	superbase->addChild(osd_switch);

	// optimalizacia grafu
	cout << "optimalizacia ..." << endl;
	optimizer.optimize(superbase);

	viewer.setUpViewer(osgProducer::Viewer::STANDARD_SETTINGS);
	viewer.getEventHandlerList().push_front(KeyOvladanie);
	viewer.setSceneData(superbase);

	viewer.setClearColor(osg::Vec4(0.7578125, 0.7265625, 0.5859375, 1.0));
	viewer.realize();
	
	while( !viewer.done() )
    {
        // wait for all cull and draw threads to complete.
        viewer.sync();

        // update the scene by traversing it with the the update visitor which will
        // call all node update callbacks and animations.
        viewer.update();
         
        // fire off the cull and draw traversals of the scene.
        viewer.frame();
        
    }
    
    // wait for all cull and draw threads to complete.
    viewer.sync();

    // run a clean up frame to delete all OpenGL objects.
    viewer.cleanup_frame();

    // wait for all the clean up frame to complete.
    viewer.sync(); 

	return 0;
}
