/**********************************************************************************/
/***	projekt:	Generator stromov, bakalarska praca							***/
/***	autor:		Tomas Pafco, xpafco00@stud.fit.vutbr.cz						***/
/***	rok:		2006/2007													***/
/***																			***/
/***	subor:		geometria.cpp												***/
/***	popis:		geometricke a matematicke vypocty 							***/
/***																			***/
/***	poznamky:	vytvorene vo Visual Studiu 2003 pod WinXP					***/
/***																			***/
/**********************************************************************************/

#include <osg/Geode>
#include <osg/Matrix>
#include <math.h>

#include "geometria.h"
#include "randomc.h"

using namespace std;

#define NAN 0x0001				// "Not A Number" 

const float pi = 3.14159265;


// ::::::::::::::::::::::::::: funkcia TPerim :::::::::::::::::::::::::::::::::::::
// obvod prstenca
// a, b - susedne body na prstenci
// pocet - pocet bodov v prstenci
//
float TPerim(osg::Vec3 & a, osg::Vec3 & b, unsigned int pocet)
{
	float vzdialenost = sqrt(	(b.x() - a.x())*(b.x() - a.x()) + 
								(b.y() - a.y())*(b.y() - a.y()) + 
								(b.z() - a.z())*(b.z() - a.z()) ); 
	return vzdialenost * pocet; 
}


// ::::::::::::::::::::::::::: funkcia AlfaComp :::::::::::::::::::::::::::::::::::
// vypocet uhlu priemetu vektoru v rovine XY
//
float AlfaComp(float dx, float dy)
{
	float DkXY = sqrt(dx*dx + dy*dy);	// dlzka priemetu
	
	// URCENIE UHLU ALFA
	//
	if ((dx == 0.0) && (dy == 0.0))
		// vektor je rovnobezny s osou z
		return NAN;

	if (dx == 0)
		if (dy > 0)
			// vektor lezi v kladnej polrovine YZ
			return pi/2;
		else
			// vektor lezi v zapornej polrovine YZ
			return (3*pi)/2;

	if (dy == 0)
		if (dx > 0)
			// vektor lezi v kladnej polrovine ZX
			return 0;
		else
			// vektor lezi v zapornej polrovine ZX
			return pi;

	if ((dx != 0.0) && (dy != 0.0))
		if (dx > 0)
			if (dy > 0)
			{	
				// 1. kvadrant
				return acos(dx/DkXY);
			}
			else
			{
				// 4. kvadrant
				return asin(dx/DkXY) + (3*pi)/2;
			}
		else
			if (dy > 0)
			{
				// 2. kvadrant
				return acos(dy/DkXY) + pi/2;
			}
			else
			{
				// 3. kvadrant
				return acos((-dx)/DkXY) + pi;
			}

	return NAN;
}

// ::::::::::::::::::::::::::: funkcia RozptylV :::::::::::::::::::::::::::::::::::
// rozptyl bodov vetvy
// x0, y0 ,z0 - suradnice predchadzajuceho bodu
// 'bod' - bod ktory sa transformuje
//
void RozptylV(osg::Vec3 & bod, float x0, float y0, float z0, float rozptyl, TRandomMersenne & rg)
{
	osg::Matrix mRotX;		// -.
	osg::Matrix mRotY;		//  |- rotacne matice
	osg::Matrix mRotZ;		// -'
	osg::Matrix mTran;		// presun do zaciatku suradnicoveho systemu
	osg::Matrix mTranB;		// presun naspat - do povodnej polohy
	osg::Vec4 hBod;			// bod v homogennych suradniciach
	
	float delta = XRandom(-rozptyl, rozptyl, rg);		// nahodny uhol odklonu od 'priameho' smeru, os x
	float beta = XRandom(-rozptyl, rozptyl, rg);		// nahodny uhol odklonu od 'priameho' smeru, os y
	float alfa = XRandom(-rozptyl, rozptyl, rg);		// nahodny uhol odklonu od 'priameho' smeru, os z
	
	float sinX = sin(delta);									
	float cosX = cos(delta);
	
	float sinY = sin(beta);									
	float cosY = cos(beta);
	
	float sinZ = sin(alfa);									
	float cosZ = cos(alfa);
										
	// nastavenie rotacnych matic
	// x
	mRotX.set(	1,		0,		0,		0,
				0,		cosX,	-sinX,	0,
				0,		sinX,	cosX,	0,
				0,		0,		0,		1);
	//y
	mRotY.set(	cosY,	0,		sinY,	0,
				0,		1,		0,		0,
				-sinY,	0,		cosY,	0,
				0,		0,		0,		1);
	//z
	mRotZ.set(	cosZ,	-sinZ,	0,		0,
				sinZ,	cosZ,	0,		0,
				0,		0,		1,		0,
				0,		0,		0,		1);

	// nastavenie translacnych matic
	mTran.set(	1,	0,	0,	-x0,
				0,	1,	0,	-y0,
				0,	0,	1,	-z0,
				0,	0,	0,	1);

	mTranB.set(	1,	0,	0,	x0,
				0,	1,	0,	y0,
				0,	0,	1,	z0,
				0,	0,	0,	1);
	
	// prepis bodu do homogennych suradnic
	hBod.set(bod.x(), bod.y(), bod.z(), 1);
	
	// presun do zaciatku suradnicoveho systemu
	hBod = mTran.postMult(hBod);

	// rotacia
	hBod = mRotX.postMult(hBod);
	hBod = mRotY.postMult(hBod);
	hBod = mRotZ.postMult(hBod);

	// presun do povodnej polohy
	hBod = mTranB.postMult(hBod);

	bod.set(hBod.x(), hBod.y(), hBod.z());
	
}


// ::::::::::::::::::::::::::: funkcia SMTrans ::::::::::::::::::::::::::::::::::::
// TRANSFORMACIA SMEROVEHO VEKTORU VETVY
// funkcia transformuje smerovy vektor vetvy, ktory bol povodne urceny vzhladom k osi z.
// vysledny (transformvoany) vektor urcuje smer vyrastania vetvy vzhladom k smerovemu
// vektoru vetvy predchadzajucej generacie
// x0, x0, z0 - pociatocny bod (k_vektoru aj smeroveho vektoru)
// dx, dy, dz - referencia na smerovy vektor
//
void SMTrans(float x0, float y0, float z0, float & dx, float & dy, float & dz, osg::Vec3 k_vektor)
{
	osg::Matrix mRotY;		// rotacne matice
	osg::Matrix mRotZ;		// 
	osg::Matrix mTran;		// presun do zaciatku suradnicoveho systemu
	osg::Matrix mTranB;		// presun naspat - do povodnej polohy
	
	// zlozky k_vektora
	float kx = k_vektor.x();
	float ky = k_vektor.y();
	float kz = k_vektor.z();

	float DkZX, DkXY;				// dlzky priemetov k_vektoru v jednotlivych rovinach
	float alfa;						// uhol k_vektoru v rovine XY
	float beta;						// uhol k_vektoru v rovine ZX
	
	osg::Vec4 A(x0 + dx, y0 + dy, z0 + dz, 1);		//
	osg::Vec4 K(x0 + kx, y0 + ky, z0 + kz, 1);		// pomocne body, homogenne suradnice
	osg::Vec4 O(x0, y0, z0, 1);						//
	
	// dlzka priemetu
	DkXY = sqrt(kx*kx + ky*ky);
	
	// URCENIE UHLU ALFA
	//
	alfa = AlfaComp(kx, ky);
	if (alfa == NAN)
		// k_vektor je rovnobezny s osou z, smer. vektor sa nebude transformovat
		return;
	
	// URCENIE UHLU BETA
	//
	// nastavenie translacnej matice
	mTran.set(	1,	0,	0,	-x0,
				0,	1,	0,	-y0,
				0,	0,	1,	-z0,
				0,	0,	0,	1);

	// presun k_vektoru do zaciatku surad. systemu
	O = mTran.postMult(O);
	K = mTran.postMult(K);

	// nastavenie rotacnej matice Z (spatna rotacia)
	mRotZ.set(	cos(alfa),	sin(alfa),	0,			0,
				-sin(alfa),	cos(alfa),	0,			0,
				0,			0,			1,			0,
				0,			0,			0,			1);

	// rotacia k_vektoru do roviny ZX
	O = mRotZ.postMult(O);
	K = mRotZ.postMult(K);

	// vypocet parametrov
	kx = K.x() - O.x();
	kz = K.z() - O.z();
	DkZX = sqrt(kz*kz + kx*kx);

	if (kz == 0)
		// k_vektor v rovine XY
		beta = pi/2;
	
	else if (kz > 0)
	{
		// horny kvadrant
		beta = acos(kz/DkZX);
	}
	else
	{
		// spodny kvadrant
		beta = acos(kx/DkZX) + pi/2;
	}

	// nastavenie bodu O do povodnej polohy
	O.set(x0, y0, z0, 1);

	// presun smeroveho vektoru do pociatku surad. systemu
	O = mTran.postMult(O);
	A = mTran.postMult(A);

	// nastavenie rotacnej matice - rotacia Y
	mRotY.set(	cos(beta),	0,		sin(beta),		0,
				0,			1,		0,				0,
				-sin(beta),	0,		cos(beta),		0,
				0,			0,		0,				1);

	// natocenie smer. vektoru podla osi Y
	O = mRotY.postMult(O);
	A = mRotY.postMult(A);

	// nastavenie rotacnej matice Z 
	mRotZ.set(	cos(alfa),	-sin(alfa),	0,			0,
				sin(alfa),	cos(alfa),	0,			0,
				0,			0,			1,			0,
				0,			0,			0,			1);

	// natocenie smer. vektoru podla osi Z
	O = mRotZ.postMult(O);
	A = mRotZ.postMult(A);
	
	// nastavenie translacnej matice
	mTranB.set(	1,	0,	0,	x0,
				0,	1,	0,	y0,
				0,	0,	1,	z0,
				0,	0,	0,	1);
	
	// presun naspat do povodnej polohy
	O = mTranB.postMult(O);
	A = mTranB.postMult(A);

	dx = A.x() - O.x();
	dy = A.y() - O.y();
	dz = A.z() - O.z();

	// *
}




// ::::::::::::::::::::::::::: funkcia GTrans :::::::::::::::::::::::::::::::::::::
// transformacia smeroveho vektoru vplyvom gravitacie a zakrivenia smerom hore
// dx, dy, dz - referencia na smerovy vektor
// uhol - uhol odklonu od osi z (smerom k rovine XY)
//
void GTrans(float & dx, float & dy, float & dz, float uhol)
{
	osg::Vec4 S(0, 0, 0, 1);		// pomocne body, homogenne suradnice
	osg::Vec4 O(0, 0, 0, 1);		//

	osg::Matrix mRotY;		// rotacne matice
	osg::Matrix mRotZ;		// 
	
	float DXY = sqrt(dx*dx + dy*dy);		// dlzka priemetu smer. vektoru v rovine XY
	float alfa;								// uhol smer. vektoru v rovine XY
	
	// URCENIE UHLU ALFA
	//
	alfa = AlfaComp(dx, dy);
	if (alfa == NAN)
		// k_vektor je rovnobezny s osou z, smer. vektor sa nebude transformovat
		return;

	// TRANSFORMACIA
	// pomocne body
	O.set(0, 0, 0, 1);
	S.set(dx, dy, dz, 1);

	// nastavenie rotacnej matice Z (spatna rotacia)
	mRotZ.set(	cos(alfa),	sin(alfa),	0,			0,
				-sin(alfa),	cos(alfa),	0,			0,
				0,			0,			1,			0,
				0,			0,			0,			1);

	// rotacia smer. vektoru do roviny ZX
	O = mRotZ.postMult(O);
	S = mRotZ.postMult(S);

	// nastavenie rotacnej matice - rotacia Y
	mRotY.set(	cos(uhol),	0,		sin(uhol),		0,
				0,			1,		0,				0,
				-sin(uhol),	0,		cos(uhol),		0,
				0,			0,		0,				1);

	// natocenie smer. vektoru podla osi Y
	O = mRotY.postMult(O);
	S = mRotY.postMult(S);

	// nastavenie rotacnej matice Z
	mRotZ.set(	cos(alfa),	-sin(alfa),	0,			0,
				sin(alfa),	cos(alfa),	0,			0,
				0,			0,			1,			0,
				0,			0,			0,			1);

	// natocenie smer. vektoru podla osi Z - naspat do povodnej polohy
	O = mRotZ.postMult(O);
	S = mRotZ.postMult(S);

	dx = S.x() - O.x();
	dy = S.y() - O.y();
	dz = S.z() - O.z();


	// *
}



// :::::::::::::::::::::::::::funkcia VertComp ::::::::::::::::::::::::::::::::::::
// vypocet suradnic povrchovych bodov
// pred - predchadzajuci kontrolny bod
// aktual - akltualny kontrolny bod v okoli ktoreho sa vypocitaju prislusne povrchove body
// next - nasledujuci kontrolny bod
// polomer v aktualnom bode
// krBody1 - pole kam sa ulozia vysledne povrchove body
// pocet - pocet povrchovych bodov v prstenci
//
void VertComp(osg::Vec3 pred, osg::Vec3 aktual, osg::Vec3 next, float polomer, 
			  osg::Vec3Array & krBody, osg::Vec3Array & normaly, int pocet)

{
	osg::Matrix mRotY;		// rotacne matice
	osg::Matrix mRotZ;		// 
	osg::Matrix mTran;		// translacna matica
	
	float dx = next.x() - pred.x();		// -
	float dy = next.y() - pred.y();		// - zlozky normaloveho vektoru
	float dz = next.z() - pred.z();		// -
	float DkZX, DkXY;					// dlzky priemetov normaloveho vektoru v jednotlivych rovinach
	float alfa;							// uhol normaloveho vektoru v rovine XY
	float beta;							// uhol normaloveho vektoru v rovine ZX
	float fi, uhol;

	int i;
	
	// dlzka priemetu
	DkXY = sqrt(dx*dx + dy*dy);
	
	// URCENIE UHLU ALFA
	// 
	alfa = AlfaComp(dx, dy);
	if (alfa == NAN)
		// norm. vektor je rovnobezny s osou z
		alfa = 0.0;
	
	// URCENIE UHLU BETA
	//
	// nastavenie translacnej matice
	mTran.set(	1,	0,	0,	-pred.x(),
				0,	1,	0,	-pred.y(),
				0,	0,	1,	-pred.z(),
				0,	0,	0,	1);

	// presun norm. vektoru do zaciatku surad. systemu
	pred = mTran.postMult(pred);
	next = mTran.postMult(next);

	// nastavenie rotacnej matice Z (spatna rotacia)
	mRotZ.set(	cos(alfa),	sin(alfa),	0,			0,
				-sin(alfa),	cos(alfa),	0,			0,
				0,			0,			1,			0,
				0,			0,			0,			1);

	// rotacia norm. vektoru do roviny ZX
	pred = mRotZ.postMult(pred);
	next = mRotZ.postMult(next);

	// vypocet parametrov
	dx = next.x() - pred.x();
	dz = next.z() - pred.z();
	DkZX = sqrt(dz*dz + dx*dx);

	if (dz == 0)
		// norm. vektor v rovine XY
		beta = pi/2;
	
	else if (dz > 0)
	{
		// horny kvadrant
		beta = acos(dz/DkZX);
	}
	else
	{
		// spodny kvadrant
		beta = acos(dx/DkZX) + pi/2;
	}

	// VYPOCET SURADNIC POVRCHOVYCH BODOV A NORMAL
	krBody.clear();
	normaly.clear();
	fi = (2*pi)/pocet;
	pocet++;						// vytvori sa o 1 bod viac, koli previazaniu textury kmena
	uhol = 0.0;
	for (i = 0; i < pocet; i++)
	{
		krBody.push_back(osg::Vec3(cos(uhol)*polomer, sin(uhol)*polomer, 0));
		normaly.push_back(osg::Vec3(cos(uhol), sin(uhol), 0));
		uhol = uhol + fi;
	}

	// nastavenie rotacnej matice Z (spatna rotacia -alfa)
	mRotZ.set(	cos(alfa),	sin(alfa),	0,			0,
				-sin(alfa),	cos(alfa),	0,			0,
				0,			0,			1,			0,
				0,			0,			0,			1);

	// 'predtocenie' bodov a normal podla osi Z
	for (i = 0; i < pocet; i++)
	{
		krBody[i] = mRotZ.postMult(krBody[i]);
		normaly[i] = mRotZ.postMult(normaly[i]);
	}
	
	
	// TRANSFORMACIA SURADNIC POVRCHOVYCH BODOV A NORMAL
	// nastavenie rotacnej matice - rotacia Y
	mRotY.set(	cos(beta),	0,		sin(beta),		0,
				0,			1,		0,				0,
				-sin(beta),	0,		cos(beta),		0,
				0,			0,		0,				1);

	// natocenie bodov a normal podla osi Y
	for (i = 0; i < pocet; i++)
	{
		krBody[i] = mRotY.postMult(krBody[i]);
		normaly[i] = mRotY.postMult(normaly[i]);
	}

	// nastavenie rotacnej matice Z 
	mRotZ.set(	cos(alfa),	-sin(alfa),	0,			0,
				sin(alfa),	cos(alfa),	0,			0,
				0,			0,			1,			0,
				0,			0,			0,			1);

	// natocenie bodov a normal podla osi Z
	for (i = 0; i < pocet; i++)
	{
		krBody[i] = mRotZ.postMult(krBody[i]);
		normaly[i] = mRotZ.postMult(normaly[i]);
	}
	
	// nastavenie translacnej matice
	mTran.set(	1,	0,	0,	aktual.x(),
				0,	1,	0,	aktual.y(),
				0,	0,	1,	aktual.z(),
				0,	0,	0,	1);
	
	// presun povrchovych bodov do prislusnej polohy
	for (i = 0; i < pocet; i++)
	{
		krBody[i] = mTran.postMult(krBody[i]);
	}

	//*
}