/**********************************************************************************/
/***	projekt:	Generator stromov, bakalarska praca							***/
/***	autor:		Tomas Pafco, xpafco00@stud.fit.vutbr.cz						***/
/***	rok:		2006/2007													***/
/***																			***/
/***	subor:		listy.cpp													***/
/***	popis:		generovanie listov - zhlukov a bilbordov					***/
/***																			***/
/***	poznamky:	vytvorene vo Visual Studiu 2003 pod WinXP					***/
/***																			***/
/**********************************************************************************/

#include <osg/Group>
#include <osg/Geode>
#include <osg/Geometry>
#include <osg/PolygonMode>
#include <osg/LightModel>
#include <osg/Material>
#include <osg/Texture2D>
#include <math.h>

#include "randomc.h"
#include "geometria.h"
#include "listy.h"
#include "tex.h"

#define alfa_kor 0.1					// korekcia rozsahov nahodnych hodnot smeroveho uhla alfa

const float pi = 3.14159265;

using namespace std;


// :::::::::::::::::::::::::: funkcia LBilbord ::::::::::::::::::::::::::::::::::::
// vytvorenie listoveho bilbordu
//
void LBilbord(osg::Group & root, TRandomMersenne & rg, osg::Vec4 start_bod, osg::Vec3 k_vektor, 
			  float uhol, float dlzka, float sirka, int kvadrant, float transluc)
{
	osg::Geode* bilGeo = new osg::Geode();
	osg::Geometry* bilbord = new osg::Geometry;
	osg::DrawElementsUInt* bilDEUI = new osg::DrawElementsUInt(osg::PrimitiveSet::QUADS, 0);

	osg::Vec3Array* vertexy = new osg::Vec3Array;	// pole vertexov
	osg::Vec3Array* normaly = new osg::Vec3Array;	// pole normal
	osg::Vec2Array* koord = new osg::Vec2Array;		// pole tex. koordinatov
	
	osg::StateSet* stSet = new osg::StateSet;
	osg::PolygonMode* pMod = new osg::PolygonMode;
	osg::LightModel* lMod = new osg::LightModel;
	osg::Material* bMat = new osg::Material;

	float alfa, beta;
	float x, y, z;
	float ux, uy, uz;		// vektor u
	float vx, vy, vz;		// vektor v, spolu s vektorom u urcuje rovinu bilbordu
	float nx, ny, nz, nd;

	beta = uhol;			// pociatocny uhol odklonenia od rodicovskej vetvy
	
	if (kvadrant == 1)
	// vetva v 1. kvadrante
		alfa = XRandom(0 + alfa_kor, pi/2 - alfa_kor, rg);
	else if (kvadrant == 2)
	// vetva v 2. kvadrante
		alfa = XRandom(pi/2 + alfa_kor, pi - alfa_kor, rg);
	else if (kvadrant == 3)
	// vetva v 3. kvadrante
		alfa = XRandom(pi + alfa_kor, (3*pi)/2 - alfa_kor, rg);
	else if (kvadrant == 4)
	// vetva v 4. kvadrante
		alfa = XRandom((3*pi)/2 + alfa_kor, 2*pi - alfa_kor, rg);
	else
	// na kvadrante nezalezi
		alfa = XRandom(0, 2*pi, rg);
	
	
	x = start_bod.x();					//
	y = start_bod.y();					//	suradnice pociatocneho bodu vetvy
	z = start_bod.z();					//
	
	vx = dlzka * sin(beta) * cos(alfa);			//
	vy = dlzka * sin(beta) * sin(alfa);			// vektor v
	vz = dlzka * cos(beta);						//

	// transformacia vektoru v
	SMTrans(x, y, z, vx, vy, vz, k_vektor);

	sirka = sirka / 2;
	beta = pi / 2;
	alfa = XRandom(0, 2*pi, rg);

	ux = sirka * sin(beta) * cos(alfa);			//
	uy = sirka * sin(beta) * sin(alfa);			// vektor u
	uz = sirka * cos(beta);						//

	// transformacia vektoru u
	SMTrans(x, y, z, ux, uy, uz, osg::Vec3(vx, vy, vz));

	// vypocet bodov, normal a koordinatov listu
	vertexy->push_back(osg::Vec3(x-ux, y-uy, z-uz));
	vertexy->push_back(osg::Vec3(x+ux, y+uy, z+uz));
	vertexy->push_back(osg::Vec3(x+ux+vx, y+uy+vy, z+uz+vz));
	vertexy->push_back(osg::Vec3(x-ux+vx, y-uy+vy, z-uz+vz));
	
	nx = uy*vz-uz*vy;
	ny = uz*vx-ux*vz;
	nz = ux*vy-uy*vx;
	nd = sqrt(nx*nx + ny*ny + nz*nz);
	nx = nx/nd;
	ny = ny/nd;
	nz = nz/nd;

	normaly->push_back(osg::Vec3(nx, ny, nz));
	koord->push_back(osg::Vec2(0, 1));
	koord->push_back(osg::Vec2(1, 1));
	koord->push_back(osg::Vec2(1, 0));
	koord->push_back(osg::Vec2(0, 0));

	// vytvorenie zobrazitelneho listu
	bilbord->setVertexArray(vertexy);
	bilbord->setNormalArray(normaly);
	bilbord->setTexCoordArray(0, koord);
	bilbord->setNormalBinding(osg::Geometry::BIND_PER_PRIMITIVE);
	
	bilDEUI->push_back(0);
	bilDEUI->push_back(1);
	bilDEUI->push_back(2);
	bilDEUI->push_back(3);
	
	// nastavenie materialu a sposobu zobrazenia
	pMod->setMode(osg::PolygonMode::Face::FRONT_AND_BACK, osg::PolygonMode::Mode::FILL);
	lMod->setTwoSided(true);
	bMat->setDiffuse(osg::Material::Face::FRONT_AND_BACK, osg::Vec4(1.0, 1.0, 1.0, 1.0));
	bMat->setEmission(osg::Material::Face::FRONT_AND_BACK, osg::Vec4(transluc, transluc, transluc, 1.0));
	
	stSet->setAttribute(pMod);
	stSet->setAttribute(lMod);
	stSet->setAttribute(bMat);
	
	// nastavenie textury
	ListTex(*stSet, rg);
	
	bilbord->addPrimitiveSet(bilDEUI);
	bilGeo->addDrawable(bilbord);
	bilGeo->setStateSet(stSet);
	root.addChild(bilGeo);

}


// ::::::::::::::::::::::::::: funkcia SphList ::::::::::::::::::::::::::::::::::::
// vytvorenie listu (sfericka metoda)
//
void SphList(osg::Group & root, float alfa, float beta, osg::Vec3 stred, float velkost, 
			 TRandomMersenne & rg)
{
	osg::Geode* listGeo = new osg::Geode();
	osg::Geometry* list = new osg::Geometry;
	osg::DrawElementsUInt* listDEUI = new osg::DrawElementsUInt(osg::PrimitiveSet::QUADS, 0);
	
	osg::StateSet* stSet = new osg::StateSet;
	osg::PolygonMode* pMod = new osg::PolygonMode;
	osg::LightModel* lMod = new osg::LightModel;
	osg::Material* listMat = new osg::Material;
	
	osg::Vec3Array* body = new osg::Vec3Array;
	osg::Vec3Array* norm = new osg::Vec3Array;
	osg::Vec2Array* coors = new osg::Vec2Array;
	osg::Matrix mRot;		// rotacna matica
	osg::Matrix mTran;		// translacna matica
	float uhol = XRandom(0, 2*pi, rg);
	bool twoSide = true;
	
	// list v strede sur. systemu
	body->push_back(osg::Vec3(-velkost/2, 0, -velkost/2));
	body->push_back(osg::Vec3(velkost/2, 0, -velkost/2));
	body->push_back(osg::Vec3(velkost/2, 0, velkost/2));
	body->push_back(osg::Vec3(-velkost/2, 0, velkost/2));
	norm->push_back(osg::Vec3(0, -1, 0));
	coors->push_back(osg::Vec2(0, 1));
	coors->push_back(osg::Vec2(1, 1));
	coors->push_back(osg::Vec2(1, 0));
	coors->push_back(osg::Vec2(0, 0));
	
	// rotacia okolo osi z o nahodny uhol
	mRot.set(	cos(uhol),	-sin(uhol),	0,			0,
				sin(uhol),	cos(uhol),	0,			0,
				0,			0,			1,			0,
				0,			0,			0,			1);
	(*body)[0] = mRot.postMult((*body)[0]);
	(*body)[1] = mRot.postMult((*body)[1]);
	(*body)[2] = mRot.postMult((*body)[2]);
	(*body)[3] = mRot.postMult((*body)[3]);
	(*norm)[0] = mRot.postMult((*norm)[0]);
	
	// rotacia okolo osi y - uhol beta
	mRot.set(	cos(beta),	0,		sin(beta),		0,
				0,			1,		0,				0,
				-sin(beta),	0,		cos(beta),		0,
				0,			0,		0,				1);
	(*body)[0] = mRot.postMult((*body)[0]);
	(*body)[1] = mRot.postMult((*body)[1]);
	(*body)[2] = mRot.postMult((*body)[2]);
	(*body)[3] = mRot.postMult((*body)[3]);
	(*norm)[0] = mRot.postMult((*norm)[0]);
	
	// rotacia okolo osi z  - uhol alfa
	mRot.set(	cos(alfa),	-sin(alfa),	0,			0,
				sin(alfa),	cos(alfa),	0,			0,
				0,			0,			1,			0,
				0,			0,			0,			1);
	(*body)[0] = mRot.postMult((*body)[0]);
	(*body)[1] = mRot.postMult((*body)[1]);
	(*body)[2] = mRot.postMult((*body)[2]);
	(*body)[3] = mRot.postMult((*body)[3]);
	(*norm)[0] = mRot.postMult((*norm)[0]);
	
	// presun do urcenej polohy
	mTran.set(	1,	0,	0,	stred.x(),
				0,	1,	0,	stred.y(),
				0,	0,	1,	stred.z(),
				0,	0,	0,	1);
	(*body)[0] = mTran.postMult((*body)[0]);
	(*body)[1] = mTran.postMult((*body)[1]);
	(*body)[2] = mTran.postMult((*body)[2]);
	(*body)[3] = mTran.postMult((*body)[3]);
	
	// vytvorenie zobrazitelneho listu
	list->setVertexArray(body);
	list->setNormalArray(norm);
	list->setTexCoordArray(0, coors);
	list->setNormalBinding(osg::Geometry::BIND_PER_PRIMITIVE);
	
	listDEUI->push_back(0);
	listDEUI->push_back(1);
	listDEUI->push_back(2);
	listDEUI->push_back(3);
	
	// nastavenie materialu a sposobu zobrazenia
	pMod->setMode(osg::PolygonMode::Face::FRONT_AND_BACK, osg::PolygonMode::Mode::FILL);
	lMod->setTwoSided(twoSide);
	listMat->setDiffuse(osg::Material::Face::FRONT_AND_BACK, osg::Vec4(1.0, 1.0, 1.0, 1.0));
	
	stSet->setAttribute(pMod);
	stSet->setAttribute(lMod);
	stSet->setAttribute(listMat);
	
	// nastavenie textury
	ListTex(*stSet, rg);
	
	root.addChild(listGeo);
	listGeo->setStateSet(stSet);
	
	list->addPrimitiveSet(listDEUI);
	listGeo->addDrawable(list);
	
	//*
}



// ::::::::::::::::::::::::::: funkcia LSphSkupina ::::::::::::::::::::::::::::::::
// vypocet suradnic a orientacie listov pomocou sferickej metody
//  - listy su nahodne rozlozene v rovnekej vzdialenost 'polomer' od stredu [x,y,z] a
//    orientovane smerom od stredu
//
void LSphSkupina(osg::Group & root, float x, float y, float z, 
				 float polomer, float pocet, float velkost, TRandomMersenne & rg)
{
	osg::Vec3 stred;
	
	int i;
	float alfa, beta;		// nahodne sfericke suradnice

	// vypocet smerovych vektorov a stredov jednotlivych listov
	for (i = 0; i < pocet; i++)
	{
		alfa = XRandom(0, 2*pi, rg);
		beta = XRandom(0, pi, rg);
		
		stred.set(
			x + polomer*sin(beta)*cos(alfa), 
			y + polomer*sin(beta)*sin(alfa), 
			z + polomer*cos(beta));
		
		// vytvorenie listu
		SphList(root, alfa, beta, stred, velkost, rg);
		
	}
	//*
}