/**********************************************************************************/
/***	projekt:	Generator stromov, bakalarska praca							***/
/***	autor:		Tomas Pafco, xpafco00@stud.fit.vutbr.cz						***/
/***	rok:		2006/2007													***/
/***																			***/
/***	subor:		povrch.cpp													***/
/***	popis:		generovanie povrchu kmena a vetvi							***/
/***																			***/
/***	poznamky:	vytvorene vo Visual Studiu 2003 pod WinXP					***/
/***																			***/
/**********************************************************************************/

#include <osg/Group>
#include <osg/Geode>
#include <osg/Geometry>
#include <stdio.h>

#include "randomc.h"
#include "povrch.h"
#include "geometria.h"

using namespace std;

// ::::::::::::::::::::::::::: funkcia TPovrch ::::::::::::::::::::::::::::::::::::
// povrch vrcholu vetvy (kmena)
//
void TPovrch(osg::Geode & geod, osg::Vec4 tBod, osg::Vec3 tNorm, 
			 osg::Vec3Array & krBody, osg::Vec3Array & norm)
{
	osg::Geometry* tGeometry = new osg::Geometry(); 
	osg::DrawElementsUInt* element = new osg::DrawElementsUInt(osg::PrimitiveSet::TRIANGLES, 0);
	osg::Vec3Array* body = new osg::Vec3Array;
	osg::Vec3Array* normaly = new osg::Vec3Array;
	int kPocet, i;

	body->insert(body->end(), krBody.begin(), krBody.end());
	normaly->insert(normaly->end(), norm.begin(), norm.end());
	body->push_back(osg::Vec3(tBod.x(), tBod.y(), tBod.z()));
	normaly->push_back(osg::Vec3(tNorm.x(), tNorm.y(), tNorm.z()));
	kPocet = krBody.getNumElements();

	tGeometry->setVertexArray(body);
	tGeometry->setNormalArray(normaly);
	tGeometry->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);

	for (i = 0; i < kPocet; i++)
	{
		if (i == kPocet-1)
		{
			// posledny element
			element->push_back(i);
			element->push_back(0);
			element->push_back(kPocet);
		}
		else
		{
			element->push_back(i);
			element->push_back(i + 1);
			element->push_back(kPocet);
		}
	}
	
	tGeometry->addPrimitiveSet(element);
	geod.addDrawable(tGeometry);
	
	//*
}


// ::::::::::::::::::::::::::: funkcia QPovrch ::::::::::::::::::::::::::::::::::::
// povrch zlozeny zo stvoruholnikov
//
void QPovrch(osg::Geode & geod, osg::Vec3Array & krBody1, osg::Vec3Array & krBody2,
			 osg::Vec3Array & normaly1, osg::Vec3Array & normaly2, float & vcor, int h_mul, int v_mul)
{
	osg::Geometry* qGeometry = new osg::Geometry();
	osg::DrawElementsUInt* element = new osg::DrawElementsUInt(osg::PrimitiveSet::QUADS, 0);
	osg::Vec3Array* body = new osg::Vec3Array;
	osg::Vec3Array* normaly = new osg::Vec3Array;
	osg::Vec2Array* coors1 = new osg::Vec2Array;		// tex. koordinaty 1. prstenca
	osg::Vec2Array* coors2 = new osg::Vec2Array;		// tex. koordinaty 2. prstenca
	int bPocet, kPocet, i;
	float u_cor, v_cor1, v_cor2;						// tex. koordinaty (zlozky)

	body->insert(body->end(), krBody1.begin(), krBody1.end());
	body->insert(body->end(), krBody2.begin(), krBody2.end());
	normaly->insert(normaly->end(), normaly1.begin(), normaly1.end());
	normaly->insert(normaly->end(), normaly2.begin(), normaly2.end());
	kPocet = krBody1.getNumElements();
	bPocet = body->getNumElements();
	
	qGeometry->setVertexArray(body);
	qGeometry->setNormalArray(normaly);
	qGeometry->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);

	v_cor1 = vcor;
	v_cor2 = vcor 
			+ 
			(v_mul *
			(sqrt(	(krBody2.front().x() - krBody1.front().x())*(krBody2.front().x() - krBody1.front().x()) + 
					(krBody2.front().y() - krBody1.front().y())*(krBody2.front().y() - krBody1.front().y()) + 
					(krBody2.front().z() - krBody1.front().z())*(krBody2.front().z() - krBody1.front().z()) ))) 
			/ 
			TPerim(krBody1[0], krBody1[1], krBody1.getNumElements());
	
	
	// zostavovanie elementov (stvoruholnikov)
	for (i = 0; i < kPocet; i++)
	{
		if (i < (kPocet - 1))
		// pocet elementov je o 1 mensi ako pocet vrcholov v prstencoch
		{
			element->push_back(i + kPocet);
			element->push_back(i + kPocet + 1);
			element->push_back(i + 1);
			element->push_back(i);
		}

		u_cor = (h_mul * ((float)i)) / ((float)kPocet);
		coors1->push_back(osg::Vec2(u_cor, v_cor1));
		coors2->push_back(osg::Vec2(u_cor, v_cor2));
	}
	
	coors1->insert(coors1->end(), coors2->begin(), coors2->end());
	qGeometry->addPrimitiveSet(element);
	qGeometry->setTexCoordArray(0, coors1);
	geod.addDrawable(qGeometry);

	vcor = v_cor2;
	//*
}


// ::::::::::::::::::::::::::: funkcia KTelo ::::::::::::::::::::::::::::::::::::::
// povrch vetvy/kmena
//
void KTelo(osg::Group & root, osg::Vec4Array & k_body, int glob_detail, int h_mul, int v_mul, int gen)
{
	
	int pocet;					// pocet povrchovych bodov v 'prstenci'
	int i = 0;					// index
	float polomer;				// polomer vetvy/kmena v urcitom bode
	float vcor = 0.0;			// vertikalny koordinat textury
	
	osg::Geode* vetvaGeode = new osg::Geode();
	osg::Vec3Array * krBody1 = new osg::Vec3Array;
	osg::Vec3Array * krBody2 = new osg::Vec3Array;
	osg::Vec3Array * normaly1 = new osg::Vec3Array;
	osg::Vec3Array * normaly2 = new osg::Vec3Array;
	osg::Vec3 normala;	// normala vrcholu
	osg::Vec3 pred;		// predchadzajuci bod
	osg::Vec3 aktual;	// aktualny bod
	osg::Vec3 next;		// nasledujuci bod
	
	
	switch (glob_detail)
	{
		case 1:
		{
			if (gen == 0)
				pocet = 6;
			else if (gen == 1)
				pocet = 5;
			else if (gen == 2)
				pocet = 3;
			else 
				pocet = 3; 
			break;
		}
		case 2:
		{
			if (gen == 0)
				pocet = 8;
			else if (gen == 1)
				pocet = 6;
			else if (gen == 2)
				pocet = 5;
			else 
				pocet = 3; 
			break;
		}
		case 3:
		{
			if (gen == 0)
				pocet = 12;
			else if (gen == 1)
				pocet = 8;
			else if (gen == 2)
				pocet = 6;
			else 
				pocet = 5; 
			break;
		}
		default:
		{
			if (gen == 0)
				pocet = 8;
			else if (gen == 1)
				pocet = 6;
			else if (gen == 2)
				pocet = 5;
			else 
				pocet = 3; 
		}
	}
	
	polomer = k_body[i].w();			// polomer zakladne (1. bod)
	
	aktual.set(k_body[i].x(), k_body[i].y(), k_body[i].z());
	next.set(k_body[i+1].x(), k_body[i+1].y(), k_body[i+1].z());
	pred.set(k_body[i].x(), k_body[i].y(), k_body[i].z());
	
	// vypocet suradnic povrchovych bodov, vysledne body sa ulozia do 'krBody1',
	// a vypocet normal, do 'normaly1'
	VertComp(pred, aktual, next, polomer, *krBody1, *normaly1, pocet);
	
	i++;	// dalsi bod
	polomer = k_body[i].w();
	
	while (polomer > 0.0)
	{	
		aktual.set(k_body[i].x(), k_body[i].y(), k_body[i].z());
		next.set(k_body[i+1].x(), k_body[i+1].y(), k_body[i+1].z());
		pred.set(k_body[i-1].x(), k_body[i-1].y(), k_body[i-1].z());
		
		// vypocet suradnic povrchovych bodov a normal
		VertComp(pred, aktual, next, polomer, *krBody2, *normaly2, pocet);

		// siet KRBODY1/KRBODY2
		QPovrch(*vetvaGeode, *krBody1, *krBody2, *normaly1, *normaly2, vcor, h_mul, v_mul);
		
		// presun prvov: KRBODY1 <- KRBODY2
		*krBody1 = *krBody2;
		*normaly1 = *normaly2;

		i++;
		polomer = k_body[i].w();
	}

	
	// spicka
	normala.set(
		k_body[i].x()-k_body[i-1].x(), 
		k_body[i].y()-k_body[i-1].y(), 
		k_body[i].z()-k_body[i-1].z());
	TPovrch(*vetvaGeode, k_body[i], normala, *krBody1, *normaly1);
	
	root.addChild(vetvaGeode);
	//*
}

